unit class Pheix::Model::Database::DBI;

use DBIish;
use Pheix::Constants;
use Pheix::Model::Database::Filechain;

has Str  $.host is default(PHEIX_PGSQL_HOST);
has Cool $.port is default(PHEIX_PGSQL_PORT);
has Str  $.user is default(PHEIX_PGSQL_USER);
has Str  $.password is default(q{});
has Str  $.database is default('pheix');
has Str  $.schema is default('public');
has Str  $.table;
has Bool $.test;
has      $.fields;
has      $.config is default({}) is rw;
has      $!dbh = try DBIish.connect('Pg', :$!host, :$!port, :$!database, :$!user, :$!password);

method try_connection returns Bool {
    return $!dbh && $!dbh.ping ?? True !! False;
}

method get_path(Str :$tab) returns Str {
    return $*CWD;
}

method get_modify_time(Str :$tab) returns Instant {
    return now;
}

method get_fields(Str :$tab, Bool :$force = False) returns List {
    my @ret;

    return @ret unless $!dbh && $!dbh.ping;

    my $sth = $!dbh.prepare(q:to/STATEMENT/);
    SELECT column_name
    FROM information_schema.columns
    WHERE table_schema = ?
    AND table_name = ?
    ;
    STATEMENT

    try {
        $sth.execute($!schema, $tab // $!table);
        @ret = $sth.allrows(:array-of-hash).map({ $_<column_name> });

        CATCH {
            default {
                my $e = .message;

                warn $e;
            }
        }
    }

    return $force ?? @ret !! $!fields // @ret;
}

method table_exists(Str :$t) returns Bool {
    my @values;
    my Bool $ret = False;

    return $ret unless $!dbh && $!dbh.ping;

    my $sth = $!dbh.prepare(q:to/STATEMENT/);
    SELECT EXISTS (
    SELECT FROM
        pg_tables
    WHERE
        schemaname = 'public' AND
        tablename  = ?
    );
    STATEMENT

    try {
        $sth.execute($t // $!table);
        @values = $sth.row;

        CATCH {
            default {
                my $e = .message;

                warn $e;
            }
        }
    }

    $ret = True if @values && @values.elems && @values.head == True;

    return $ret;
}

method table_create(Str :$t) returns Hash {
    my %ret = status => False;

    return %ret unless $!fields && $!fields.elems && $!dbh && $!dbh.ping;

    my Str $q_fields = $!fields.map({ sprintf("\"%s\" %s", $_, $_ ~~ /id/ ?? 'SERIAL' !! 'TEXT')}).join(q{,});

    try {
        $!dbh.execute(sprintf("CREATE TABLE \"%s\" (%s);", $t // $!table, $q_fields));

        %ret<status> = True;

        CATCH {
            default {
                %ret<error>  = .message;
                %ret<status> = False;

                warn %ret<error>;
            }
        }
    }

    return %ret;
}

method table_drop(Str :$t) returns Bool {
    my Bool $ret = False;

    return $ret unless $!dbh && $!dbh.ping;

    try {
        $!dbh.execute(sprintf("DROP TABLE \"%s\";", $t // $!table));

        $ret = True;

        CATCH {
            default {
                my $e = .message;

                warn $e;
            }
        }
    }

    return True;
}

method table_raw returns List {
    my @rows;

    return @rows unless $!dbh && $!dbh.ping;

    try {
        my $sth = $!dbh.execute(sprintf("SELECT * FROM \"%s\" LIMIT 500;", $!table));
        @rows   = $sth.allrows(:array-of-hash);

        CATCH {
            default {
                my $e = .message;

                warn $e;
            }
        }
    }

    return @rows;
}

method row_insert(
    Str  :$t,
    UInt :$id,
    Hash :$data!,
) returns Hash {
    my %ret = status => False;

    return %ret unless $!dbh && $!dbh.ping && $data.keys;

    my $q_fields = $data.keys.sort.map({ sprintf("\"%s\"", $_) }).join(q{,});
    my $q_values = $data
        .keys
        .sort
        .map({
            $_ eq 'id' && $data{$_} ~~ /^[0..9]+$/ ??
                $data{$_} !!
                    sprintf("'%s'", $data{$_})
        })
        .join(q{,});

    try {
        $!dbh.execute(sprintf("INSERT INTO \"%s\" (%s) VALUES (%s);", $t // $!table, $q_fields, $q_values));

        CATCH {
            default {
                %ret<error>  = .message;
                %ret<status> = False;

                warn %ret<error>;
            }
        }
    }

    %ret<status> = True;

    return %ret;
}

method row_set(
    Str  :$t,
    Hash :$data!,
    Hash :$clause!
) returns Hash {
    my %ret = status => False;

    return %ret unless $!dbh && $!dbh.ping && $data.keys && $clause.keys;

    my $update = $data
        .keys
        .sort
        .map({
            $_ eq 'id' && $data{$_} ~~ /^[0..9]+$/ ??
                sprintf("\"%s\" = %d", $_, $data{$_}) !!
                    sprintf("\"%s\" = '%s'", $_, $data{$_})
        })
        .join(q{,});

    my $where = self!hash_to_where(:$clause);

    try {
        $!dbh.execute(sprintf("UPDATE \"%s\" SET %s WHERE %s;", $t // $!table, $update, $where));
        %ret<status> = True;

        CATCH {
            default {
                %ret<error>  = .message;
                %ret<status> = False;

                warn %ret<error>;
            }
        }
    }

    return %ret;
}

method row_get(Str :$t, Hash :$clause!) returns List {
    my @rows;

    return @rows unless $!dbh && $!dbh.ping && $clause.keys;

    my $where = self!hash_to_where(:$clause);

    try {
        my $sth = $!dbh.execute(sprintf("SELECT * FROM \"%s\" WHERE %s LIMIT 500;", $t // $!table, $where));
        @rows   = $sth.allrows(:array-of-hash);

        CATCH {
            default {
                my $e = .message;

                warn $e;
            }
        }
    }

    return @rows;
}

method row_remove(Str :$t, Hash :$clause!) returns Bool {
    my Bool $ret = False;

    my $where = self!hash_to_where(:$clause);

    try {
        my $sth = $!dbh.execute(sprintf("DELETE FROM \"%s\" WHERE %s;", $t // $!table, $where));
        $ret = True;

        CATCH {
            default {
                my $e = .message;

                warn $e;
            }
        }
    }

    return $ret;
}

method get_from_array(Hash $arg_ref, @lines, Bool $bchain?) returns List {
    return Pheix::Model::Database::Filechain.new.get_from_array($arg_ref, @lines, $bchain);
}

method !hash_to_where(Hash :$clause!) returns Str {
    return q{} unless $clause.keys;

    return $clause
        .keys
        .map({
            $_ eq 'id' && $clause{$_} ~~ /^[0..9]+$/ ??
                sprintf("\"%s\" = %d", $_, $clause{$_}) !!
                    sprintf("\"%s\" = '%s'", $_, $clause{$_})
        })
        .join(' AND ');
}
